﻿using System.Collections.Generic;
using System.Linq;

using LogicModule.Nodes.Helpers;
using LogicModule.ObjectModel;
using LogicModule.ObjectModel.TypeSystem;

namespace LogicNodesSDK.Logic.Examples
{
  /// <summary>
  /// A code example using a variable number of inputs.
  /// Uses aggrgate functions on multiple inputs to get the minimum, maximum, sum and average of all input values.
  /// </summary>
  public class Aggregation : LocalizablePrefixLogicNodeBase
  {
    /// <summary>
    /// The prefix to use for the names of the inputs. An index is added afterwards to give a unique name.
    /// </summary>
    private const string InputPrefix = "Input";

    /// <summary>
    /// The TypeService is saved to allow updating after the constructor is finished.
    /// </summary>
    private readonly ITypeService typeService;

    /// <summary>
    /// Initializes a new instance of the <see cref="BasicNode"/> class.
    /// </summary>
    /// <param name="context">The node context.</param>
    public Aggregation(INodeContext context)
      : base(context, InputPrefix)
    {
      context.ThrowIfNull("context");

      this.typeService = context.GetService<ITypeService>();

      this.Inputs = new List<DoubleValueObject>();

      this.InputCount = this.typeService.CreateInt(PortTypes.Integer, "InputCount", 2);
      // define limits for the input count
      this.InputCount.MinValue = 2;
      this.InputCount.MaxValue = 99;

      // A helper function that grows/shrinks the list of inputs whenever the input count is changed
      ListHelpers.ConnectListToCounter(this.Inputs, this.InputCount, this.typeService.GetValueObjectCreator(PortTypes.Number, InputPrefix), null, null);

      // Initialize the outputs
      this.OutputMin = this.typeService.CreateDouble(PortTypes.Number, "OutputMin");
      this.OutputMax = this.typeService.CreateDouble(PortTypes.Number, "OutputMax");
      this.OutputSum = this.typeService.CreateDouble(PortTypes.Number, "OutputSum");
      this.OutputAverage = this.typeService.CreateDouble(PortTypes.Number, "OutputAverage");
    }



    /// <summary>
    /// A list of double value objects as inputs. The attribute <see cref="PortAttribute.IsRequired"/> is set to true, so the inputs cannot be disabled. 
    /// </summary>
    [Input(DisplayOrder = 1, InitOrder = 2, IsRequired = true)]
    public IList<DoubleValueObject> Inputs { get; private set; }

    /// <summary>
    /// How many inputs shall be used. The limits are set from 2-99 in the constructor of the <see cref="Aggregation"/> class.
    /// The attribute <see cref="PortAttribute.IsDefaultShown"/> is set to false, so this field isn't shown on the node tile in the logic editor by default.
    /// </summary>
    [Parameter(DisplayOrder = 2, InitOrder = 1, IsDefaultShown = false)]
    public IntValueObject InputCount { get; private set; }

    /// <summary>
    /// The minimum value of all input values.
    /// </summary>
    [Output(DisplayOrder = 1)]
    public DoubleValueObject OutputMin { get; private set; }

    /// <summary>
    /// The maximum value of all input values.
    /// </summary>
    [Output(DisplayOrder = 2)]
    public DoubleValueObject OutputMax { get; private set; }

    /// <summary>
    /// The sum of all input values.
    /// </summary>
    [Output(DisplayOrder = 3)]
    public DoubleValueObject OutputSum { get; private set; }

    /// <summary>
    /// The average value of all input values.
    /// </summary>
    [Output(DisplayOrder = 4)]
    public DoubleValueObject OutputAverage { get; private set; }

    public override void Execute()
    {
      // calculate the outputs with various aggregate functions over all input values
      this.OutputMin.Value = this.Inputs.Min(valueObject => valueObject.Value);
      this.OutputMax.Value = this.Inputs.Max(valueObject => valueObject.Value);
      this.OutputSum.Value = this.Inputs.Sum(valueObject => valueObject.Value);
      this.OutputAverage.Value = this.Inputs.Average(valueObject => valueObject.Value);
    }

    // The following function is used in "LocalizablePrefixLogicNodeBase" to also translate the prefixed inputs.
    /*
    public override string Localize(string language, string key)
    {
      language.ThrowIfNull("language");
      key.ThrowIfNull("key");

      CultureInfo cultureInfo;
      try
      {
        cultureInfo = CultureInfo.GetCultureInfo(language);
      }
      catch (CultureNotFoundException)
      {
        cultureInfo = CultureInfo.InvariantCulture;
      }

      return this.ResourceManager.GetString(key, cultureInfo) ??
        this.ResourceManager.LocalizePrefix(cultureInfo, InputPrefix, key) ??
        key;
    }
    */
  }
}
