﻿using System;
using System.Collections.Generic;
using System.Linq;
using LogicModule.Nodes.Helpers;
using LogicModule.ObjectModel;
using LogicModule.ObjectModel.TypeSystem;

namespace LogicNodesSDK.Logic.Examples
{
  /// <summary>
  /// A code example using the ISchedulerService to work asynchronously.
  /// Calculates the average of input values over a specified timespan.
  /// </summary>
  public class TimedAverage : LogicNodeBase
  {
    /// <summary>
    /// The scheduler service to access the current time and perform actions in the future.
    /// </summary>
    private readonly ISchedulerService schedulerService;

    /// <summary>
    /// A token containing the timepoint when to next update the list automatically, if enabled.
    /// </summary>
    private SchedulerToken updateToken;

    /// <summary>
    /// The maximum number of values that shall be used to calculate the average.
    /// </summary>
    private const int MaxQueueSize = 100;

    /// <summary>
    /// A container containing the values and timestamps of the input over the specified time.
    /// </summary>
    private Queue<Tuple<DateTime, double>> lastValues;

    /// <summary>
    /// Initializes a new instance of the <see cref="TimedAverage"/> class.
    /// </summary>
    /// <param name="context">The node context.</param>
    public TimedAverage(INodeContext context)
      : base(context)
    {
      context.ThrowIfNull("context");

      this.schedulerService = context.GetService<ISchedulerService>();

      var typeService = context.GetService<ITypeService>();
      this.Input = typeService.CreateDouble(PortTypes.Number, "Input");
      this.Output = typeService.CreateDouble(PortTypes.Number, "Output");

      this.Reset = typeService.CreateBool(PortTypes.Binary, "Reset", false);

      // use one minute as default value
      this.AverageTime = typeService.CreateTimeSpan(PortTypes.TimeSpan, "AverageTime", new TimeSpan(0, 0, 1, 0));
      // set minimum value to one second
      this.AverageTime.MinValue = new TimeSpan(0, 0, 0, 1);

      this.UpdateAutomatically = typeService.CreateBool(PortTypes.Bool, "UpdateAutomatically", false);

      this.lastValues = new Queue<Tuple<DateTime, double>>();
    }

    /// <summary>
    /// The input to calculate the average over time of.
    /// The attribute <see cref="PortAttribute.IsRequired"/> is set to true, so it is not possible to disable this input. 
    /// </summary>
    [Input(DisplayOrder = 1, IsRequired = true)]
    public DoubleValueObject Input { get; private set; }

    /// <summary>
    /// Input to manually clear the list of values.
    /// The attributes <see cref="PortAttribute.IsDefaultShown"/> and <see cref="InputAttribute.IsInput"/> are both set to false, 
    /// so the input is not shown and not enabled by default.
    /// </summary>
    [Input(DisplayOrder = 2, IsDefaultShown = false, IsInput = false)]
    public BoolValueObject Reset { get; private set; }

    /// <summary>
    /// The timespan to average the input values over.
    /// </summary>
    [Parameter(DisplayOrder = 10)]
    public TimeSpanValueObject AverageTime { get; private set; }

    /// <summary>
    /// Whether to update the output automatically when a value gets too old and to be included in the average.
    /// </summary>
    [Parameter(DisplayOrder = 11, IsDefaultShown = false)]
    public BoolValueObject UpdateAutomatically { get; private set; }

    /// <summary>
    /// The average of the input values averaged over the specified timespan.
    /// </summary>
    [Output(DisplayOrder = 1)]
    public DoubleValueObject Output { get; private set; }

    /// <summary>
    /// Triggered when the input receives a new value or the reset input is set.
    /// </summary>
    public override void Execute()
    {
      // If the Reset was set and the value is true, clear the queue
      if (this.Reset.WasSet && this.Reset.Value)
      {
        Unschedule();
        this.lastValues.Clear();
      }

      // If the Input was set, add the new value to the queue, remove values that are too old and update the outputs
      if (this.Input.WasSet)
      {
        this.lastValues.Enqueue(new Tuple<DateTime, double>(this.schedulerService.Now.ToUniversalTime(), this.Input.Value));
        UpdateQueue();
        UpdateOutputs();
        Schedule();
      }
    }

    /// <summary>
    /// Remove entries from the queue that are too old to be used in the average and trim queue to max size
    /// </summary>
    private void UpdateQueue()
    {
      var latestTimepoint = this.schedulerService.Now.ToUniversalTime() - this.AverageTime.Value;
      while (((this.lastValues.Count > 0) && (this.lastValues.Peek().Item1 < latestTimepoint)) || (this.lastValues.Count > MaxQueueSize))
      {
        this.lastValues.Dequeue();
      }
    }

    /// <summary>
    /// Set the output to the average of the last values, if queue contains any value
    /// </summary>
    private void UpdateOutputs()
    {
      if (this.lastValues.Count > 0)
      {
        this.Output.Value = this.lastValues.Average(x => x.Item2);
      }
    }

    /// <summary>
    /// Update the queue, the outputs and schedule the next trigger execution.
    /// </summary>
    private void Trigger()
    {
      Unschedule();
      UpdateQueue();
      UpdateOutputs();
      Schedule();
    }

    /// <summary>
    /// If the output should be updated automatically, add the next trigger time to the scheduler service.
    /// </summary>
    private void Schedule()
    {
      if (this.UpdateAutomatically.HasValue && this.UpdateAutomatically.Value && (this.lastValues.Count > 0) && (this.updateToken == null))
      {
        var nextTriggerTime = this.lastValues.Peek().Item1 + this.AverageTime.Value;
        this.updateToken = this.schedulerService.InvokeAt(nextTriggerTime, this.Trigger);
      }
    }

    /// <summary>
    /// Unschedule the next trigger execution, if scheduled.
    /// </summary>
    private void Unschedule()
    {
      if (this.updateToken != null)
      {
        schedulerService.Remove(this.updateToken);
        this.updateToken = null;
      }
    }
  }
}
