﻿using LogicModule.Nodes.Helpers;
using LogicModule.ObjectModel;
using LogicModule.ObjectModel.TypeSystem;

namespace LogicNodesSDK.Logic.Examples
{
  /// <summary>
  /// A code example using a single input and output to be used as a converter on a node output.
  /// Toggles the output value between 1 and 0 whenever an input value is received.
  /// This node can be added to an output of a node by rightclicking the output of any logic node.
  /// </summary>
  public class Toggle : LogicNodeBase
  {
    /// <summary>
    /// The value to send when the node is executed next.
    /// </summary>
    bool toggleValue = true;

    /// <summary>
    /// Initializes a new instance of the <see cref="Toggle"/> class.
    /// </summary>
    /// <param name="context">The node context.</param>
    public Toggle(INodeContext context)
      : base(context)
    {
      context.ThrowIfNull("context");

      var typeService = context.GetService<ITypeService>();

      this.Input = typeService.CreateAny(PortTypes.Any, "Input");
      this.Output = typeService.CreateBool(PortTypes.Bool, "Output");
    }

    /// <summary>
    /// Use any value type as input.
    /// The attribute <see cref="PortAttribute.IsRequired"/> is set to true, so the input cannot be disabled.
    /// </summary>
    [Input(IsRequired = true)]
    public AnyValueObject Input { get; private set; }

    /// <summary>
    /// Output toggles between 1 and 0.
    /// </summary>
    [Output]
    public BoolValueObject Output { get; private set; }

    /// <summary>
    /// Set the output value to 1 or 0, then toggle the next value to send.
    /// </summary>
    public override void Execute()
    {
      this.Output.Value = toggleValue;
      toggleValue = !toggleValue;
    }
  }
}
